#!/tvbin/tivosh

# TuikHelper.tcl
# Script By TivoWare
# Version 0.2 08/30/2005 - Changed how internal commands are used.
#                        - Added function to check holder references.
# Version 0.1 08/07/2005 - First Release

set TuikHelperVersion 1.0

# This is a collection of functions useful for working with TuikResources
# Use the following code to use them in your scripts

##Require TuikHelper.tcl
#set SourceTuikHelper 1
#foreach TuikHelperPath [split "[file dirname [info script]]:.:$env(PATH)" ":"] {
#  if { [file exists "${TuikHelperPath}/TuikHelper.tcl"] == 1 } {
#    source "${TuikHelperPath}/TuikHelper.tcl"
#    break
#  }
#}
#if {[info exists TuikHelperVersion]==0} {puts "TuikHelper.tcl not found!";exit}

proc GetTuikResourceGroups {Path} {
  if { $Path == "" } { return "" } 
  set db [dbopen]
  RetryTransaction {
    set SwSystem [db $db open $Path ]
    set TuikResourceStateTemplate [dbobj $SwSystem get TuikResourceStateTemplate]
    foreach TuikResourceGroup [dbobj $TuikResourceStateTemplate gettarget TuikResourceGroup] {
      lappend TuikResourceGroupList $TuikResourceGroup
    }
  }
  dbclose $db
  return $TuikResourceGroupList
}

proc GetTuikResourceHolders {Path} {
  if { $Path == "" } { return "" } 
  set TuikResourceGroupList [GetTuikResourceGroups $Path]
  set db [dbopen]
  foreach TuikResourceGroup $TuikResourceGroupList {
    RetryTransaction {
      regexp {(.*)/(.*)} $TuikResourceGroup -> id subid
      set Group [db $db openidconstruction $id $subid]
      foreach TuikResourceHolder [dbobj $Group gettarget TuikResourceHolder] {
        lappend TuikResourceHolderList $TuikResourceHolder
      }
    }
  }
  dbclose $db
  return $TuikResourceHolderList
}

proc GetResourceList {Path args} {
  if { $Path == "" } { return "" } 
  set WithIndex 0
  if { [lindex $args 0] == 1 } { set WithIndex 1 }

  set TuikResourceHolderList [GetTuikResourceHolders $Path]
  set db [dbopen]
  foreach TuikResourceHolder $TuikResourceHolderList {
    RetryTransaction {
      regexp {(.*)/(.*)} $TuikResourceHolder -> id subid
      set Holder [db $db openidconstruction $id $subid]
      if { $WithIndex == 1 } {
        lappend TuikResources "[dbobj $Holder get PathBase] [lindex [dbobj $Holder get IndexPath] 0]"
      } else {
        lappend TuikResources [dbobj $Holder get PathBase]
      }
    }
  }
  dbclose $db
  return $TuikResources
}

proc CheckResourceHolder {Path TuikResourceFix Quiet} {
  if { $Path == "" } { return "" } 

  set TuikResourceHolderList [GetTuikResourceHolders $Path]
  set db [dbopen]
  foreach TuikResourceHolder $TuikResourceHolderList {
    RetryTransaction {
      regexp {(.*)/(.*)} $TuikResourceHolder -> id subid
      set Holder [db $db openidconstruction $id $subid]
      set Resource [dbobj $Holder get TuikResource]
      set HolderFileId [dbobj $Holder get File]
      set ResourceFileId [dbobj $Resource get File]
      if {$HolderFileId != $ResourceFileId} {
        if {$Quiet == 0 } {
          puts -nonewline "[dbobj $Holder get PathBase] holder has invalid reference!"
        }
        if {$TuikResourceFix == 1} {
          dbobj $Holder set TuikResource $Resource
          puts " - Fixed"  
        } else {
          puts ""
        }
      }
    }
  }
  dbclose $db
  return 0
}

proc FindResource {Path FindName} {
  if { $Path == "" || $FindName == "" } { return "" } 
  if { [regexp {(.*)/(.*)} $FindName] } { 
    set RegSearch 0 
  } else {
    set RegSearch 1
  }
  set TuikResourceHolderList [GetTuikResourceHolders $Path]
  set db [dbopen]
  foreach TuikResourceHolder $TuikResourceHolderList {
    RetryTransaction {
      regexp {(.*)/(.*)} $TuikResourceHolder -> id subid
      set Holder [db $db openidconstruction $id $subid]
      set PathBase [dbobj $Holder get PathBase]
      set FileBase $PathBase
      regexp {.*/(.*)} $PathBase -> FileBase
      if { $PathBase == $FindName || ($RegSearch == 1 && $FileBase == $FindName)} {
        lappend ResourceInfo $FileBase
        lappend ResourceInfo $PathBase
        lappend ResourceInfo [lindex [dbobj $Holder get IndexPath] 0]
        lappend ResourceInfo $id
        lappend ResourceInfo $subid
        dbclose $db
        return $ResourceInfo
      }
    }
  }
  dbclose $db
  return ""
}

proc FindActiveResource {FindName} {
  return [FindResource "/SwSystem/ACTIVE" $FindName]
}

proc FindResourceIndexPath {ResourceName} {
  return [lindex [FindActiveResource $ResourceName] 2]
}

proc ExportResourceByName {ResourceName FileName} {
  set ResourceIndexPath [FindResourceIndexPath $ResourceName]
  if { $ResourceIndexPath == "" } { return "" }
  return [ ExportResourceByIndexPath $ResourceIndexPath $FileName ]
}

proc ImportResourceByName {ResourceName FilePath} {
  set ResourceIndexPath [FindResourceIndexPath $ResourceName]
  if { $ResourceIndexPath == "" } { return "" }
  return [ ImportResourceByIndexPath $ResourceIndexPath $FileName ]
}

proc ExportResourceByIndexPath {ResourceIndexPath FileName} {
  if { $ResourceIndexPath == "" || $FileName == "" } { return "" }
  set db [dbopen]
  RetryTransaction {
    set TuikResourceHolder [db $db open $ResourceIndexPath ]
    set TuikResource [dbobj $TuikResourceHolder get TuikResource]
    FromMfs $FileName [dbobj $TuikResource get File]
  }
  dbclose $db
  return $FileName
}

proc ImportResourceByIndexPath {ResourceIndexPath FileName} {
  if { $ResourceIndexPath == "" || $FileName == "" } { return "" }
  set db [dbopen]
  RetryTransaction {
    set TuikResourceHolder [db $db open $ResourceIndexPath ]
    set TuikResource [dbobj $TuikResourceHolder get TuikResource]
    set NewResourceFile [ToMfs $FileName]
    dbobj $TuikResource set File $NewResourceFile
    dbobj $TuikResourceHolder set TuikResource $TuikResource
  }
  dbclose $db
  return $FileName
}


##########################################
# Don't use any functions below this line
##########################################

# Section for Built in Commands 

proc DoHelp {TuikHelperVersion} {
    puts "------------------------------------------------------------------------------"
    puts "TuikHelper.tcl by TivoWare - Version $TuikHelperVersion"
    puts "------------------------------------------------------------------------------"
    puts "Builtin Commands:   {required} \[optional\]"
    puts "  export {ResourceName} \[FileName\]  Export a TuikResource to a file."
    puts "  import {ResourceName} \[FileName\]  Import a TuikResource from a file."
    puts "  dump \[Path\] \[-brfonly\] \[-nodir\]   Dump all Resources to a directory."
    puts "  list \[FileName\] \[-withindex\]      List all Resources." 
    puts "  check  \[-fix\]                     Check TuikResourceHolder file ids."
    puts "Parameter Help:"
    puts "  {ResourceName} - Resource PathBase. Best is the full PathBase like so"
    puts "                   ui/menu/MainMenu.brf, but just MainMenu.brf can be used."
    puts "  \[FileName\]     - File to read or write data to. Default is resource name."
    puts "                   (The list command will output to the screen by default."
    puts "  -brfonly         Only extract .brf files."
    puts "  -nodir           Do not build the directory structure."
    puts "  -withindex       Output the IndexPath also."
    puts "Example:"
    puts "  TuikHelper.tcl list"
    puts "  TuikHelper.tcl export lib/util/MessagesCommon.brf"
    puts "  TuikHelper.tcl import lib/util/MessagesCommon.brf MyMessages.brf"
    puts "  TuikHelper.tcl dump brf/ -nodir -brfonly"
    puts "  TuikHelper.tcl check -fix"
    return
}

proc DoCheck {TuikHelperFix TuikHelperSwSystem} {
  CheckResourceHolder $TuikHelperSwSystem $TuikHelperFix 0
}

proc DoList {FileName TuikHelperWithIndex TuikHelperSwSystem} {
  set ResourceList [GetResourceList $TuikHelperSwSystem $TuikHelperWithIndex]
  if { $FileName != "" } {
    set FileId [ open $FileName w ]
    foreach Resource $ResourceList {
      puts $FileId $Resource
    }
    close $FileId
  } else {
    foreach Resource $ResourceList {
      puts $Resource
    }
  }
  return
}

proc DoExport {ResourceName FileName TuikHelperSwSystem} {
  if { $ResourceName == "" } {
    puts "ResourceName is required!"
    DoHelp $TuikHelperVersion
    return 
  }
  set ResourceInfo [FindResource $TuikHelperSwSystem $ResourceName]
  if { $ResourceInfo == "" } {
    puts "Failed to find $ResourceName!"
    exit
  }
  if { $FileName == "" } {
    set FileName [lindex $ResourceInfo 0]
  }
  set ResourceIndexPath [lindex $ResourceInfo 2]
  if { [ExportResourceByIndexPath $ResourceIndexPath $FileName] == "" } {
    puts "Failed to export $ResourceName!"
  }
}

proc DoImport {ResourceName FileName TuikHelperSwSystem} {
  if { $ResourceName == "" } {
    puts "ResourceName is required!"
    DoHelp $TuikHelperVersion
    return 
  }
  set ResourceInfo [FindResource $TuikHelperSwSystem $ResourceName]
  if { $ResourceInfo == "" } {
    puts "Failed to find $ResourceName!"
    exit
  }
  if { $FileName == "" } {
    set FileName [lindex $ResourceInfo 0]
  }
  if {[file exists $FileName] == 0} {
     puts "File $FileName not found!"
     exit
  }
  set ResourceIndexPath [lindex $ResourceInfo 2]
  if { [ImportResourceByIndexPath $ResourceIndexPath $FileName] == "" } {
    puts "Failed to import $ResourceName!"
  }
}

proc DoDump {Path TuikHelperBrfOnly TuikHelperNoDir TuikHelperSwSystem} {
  if { $Path != "" && [string index $Path [expr [string length $Path]-1]] != "/" } {
    set Path "${Path}/"
  }
  file mkdir $Path
  set ResourceList [GetResourceList $TuikHelperSwSystem 1]
  foreach Resource $ResourceList {
    set ResourceName ""
    set ResourceIndexPath ""
    regexp {(.*)\ (.*)} $Resource -> ResourceName ResourceIndexPath
    set FileName $ResourceName
    set FilePath ""
    if { $TuikHelperNoDir == 1 } {
      regexp {.*/(.*)} $FileName -> FileName
    } else {
      regexp {(.*)/.*} $FileName -> FilePath
      file mkdir $Path$FilePath
    }
    set Skip 0
    if { $TuikHelperBrfOnly == 1 } {
      set BrfCheck ""
      regexp {.*\.(.*)} $FileName -> BrfCheck
      if { $BrfCheck != "brf" } { set Skip 1 }
    }
    if { $Skip == 0 } {
      puts "Exporting: $FileName"
      if { [ExportResourceByIndexPath $ResourceIndexPath $Path$FileName] == "" } {
        puts "Failed to export $ResourceName!"
      }
    } 
  }
  return
}

# Main Sub
# Only do anything if SourceTuikHelper = 0

if {[info exists SourceTuikHelper] != 1} {
  set SourceTuikHelper 0
}

if { $SourceTuikHelper != 1} {
  #Process flags
  set TuikHelperBrfOnly 0
  set TuikHelperNoDir 0
  set TuikHelperWithIndex 0
  set TuikHelperFix 0
  set TuikHelperSwSystem "/SwSystem/ACTIVE"

  foreach arg $argv {
    if { $arg == "-brfonly" } { set TuikHelperBrfOnly 1 
    } elseif { $arg == "-nodir" } { set TuikHelperNoDir 1 
    } elseif { $arg == "-withindex" } { set TuikHelperWithIndex 1 
    } elseif { [string range $arg 0 9] == "-SwSystem=" } { regexp {\-SwSystem\=(.*)} $arg -> TuikHelperSwSystem  
    } elseif { $arg == "-fix" } { set TuikHelperFix 1
    } elseif { [string index $arg 0] == "-" } {
      puts "Unknown flag $arg"
      exit
    }
  }
  if { [lindex $argv 0] == "export" } {
    set ResourceName ""
    set FileName ""
    set argindex 1
    while { $argindex < $argc } {
      if { [string index [lindex $argv $argindex] 0] != "-" } {
        if { $ResourceName == "" } {
          set ResourceName [lindex $argv $argindex] 
        } else {
          set FileName [lindex $argv $argindex]
          break
        }
      }
      incr argindex
    }
    DoExport $ResourceName $FileName $TuikHelperSwSystem
    exit
  } elseif { [lindex $argv 0] == "import" } {
    set ResourceName ""
    set FileName ""
    set argindex 1
    while { $argindex < $argc } {
      if { [string index [lindex $argv $argindex] 0] != "-" } {
        if { $ResourceName == "" } {
          set ResourceName [lindex $argv $argindex] 
        } else {
          set FileName [lindex $argv $argindex]
          break
        }
      }
      incr argindex
    }
    DoImport $ResourceName $FileName $TuikHelperSwSystem
    exit
  } elseif { [lindex $argv 0] == "dump" } {
    set Path ""
    set argindex 1
    while { $argindex < $argc } {
      if { [string index [lindex $argv $argindex] 0] != "-" } {
        set Path [lindex $argv $argindex] 
        break
      }
      incr argindex
    }
    DoDump $Path $TuikHelperBrfOnly $TuikHelperNoDir $TuikHelperSwSystem
    exit
  } elseif { [lindex $argv 0] == "list" } {
    set FileName ""
    set argindex 1
    while { $argindex < $argc } {
      if { [string index [lindex $argv $argindex] 0] != "-" } {
        set FileName [lindex $argv $argindex] 
        break
      }
      incr argindex
    }
    DoList $FileName $TuikHelperWithIndex $TuikHelperSwSystem
    exit
  } elseif { [lindex $argv 0] == "check" } {
    DoCheck $TuikHelperFix $TuikHelperSwSystem
    exit
  } else {
    DoHelp $TuikHelperVersion
  }
}
