#!/tvbin/tivosh
#Copyright 2006 AlphaWolf_HK
#Neither this file nor any derivative may be distributed anywhere outside of
#the forums located at www.dealdatabase.com.
#
#Fakecall.tcl rewritten from scratch
#
#The old one was a bit convoluted due to all of the exceptions for legacy
#sofware versions. But with the release of 6.3 for hdtivos, no S2 tivos should
#be using any version below 6.x.
#
#New features include the ability to offset the next call date by an arbitrary
#amount, and a fix for a bug that caused the next call time to be wrong.
#
#To offset the next call date, just use the number of days (from today) that
#you want. e.g. to set it 3 weeks from today:
#
#fakecall.tcl 21
#
#Default is 7 days.
#

set ncoffset [lindex $argv 0]
if { $ncoffset == "" } { set ncoffset 7 }

puts "Fakecall.tcl by AlphaWolf_HK"

EnableTransactionHoldoff true

RetryTransaction {
    set db [dbopen]

    set now [clock seconds]
    set onedayinsec [expr (60 * 60 * 24)]
    set today [expr ($now / $onedayinsec)]

    #Tivosh strips off the decimal from $today, so the result is nonzero since
    #a partial day isn't included.
    set timeinsectoday [expr ($now - ($today * $onedayinsec))]
    set nextcalldate [expr ($today + $ncoffset)]
    set ssexpiration [expr ($nextcalldate * $onedayinsec)]

    set pcobj [db $db open /State/PhoneConfig]

    dbobj $pcobj set CallStatusInfo 1
    dbobj $pcobj set LastCallAttemptDay $today
    dbobj $pcobj set LastCallAttemptSecInDay $timeinsectoday
    dbobj $pcobj set LastCallStatus "Succeeded"
    dbobj $pcobj set LastDialInUpdateDate $today
    dbobj $pcobj set LastSuccessCallDay $today
    dbobj $pcobj set LastSuccessCallSecInDay $timeinsectoday
    dbobj $pcobj set LastSuccessPgdCallDay $today
    dbobj $pcobj set LastSuccessPgdCallSecInDay $timeinsectoday
    dbobj $pcobj set NextCallAttemptDays $nextcalldate
    #Offset seconds from NextCallAttemptDays, just set it to the current time.
    dbobj $pcobj set NextCallAttemptSecInDay $timeinsectoday

    #Should this be in seconds from now? 6.3a doesn't set this anymore during
    #the call in so I am not even sure that it is necessary anymore.
    set scobj [db $db open /State/ServiceConfig]
    dbobj $scobj set ServiceStateExpiration $ssexpiration

    #6.3 doesn't need this, but do 6.2 or 7.x versions need it?
    set gcobj [db $db open /State/GeneralConfig]
    try { dbobj $ssobj set Complete 7 } catch junk { }

}

puts "\"Call in\" status updated."
