#!/tvbin/tivosh
#Copyright 2006 AlphaWolf_HK
#Neither this file nor any derivative may be distributed anywhere outside of
#the forums located at www.dealdatabase.com.
#
#recstatus.tcl, written from scratch by AlphaWolf_HK 2006
#
#This script traverses /Recording/InProgress and /Recording/Active and checks
#if a recording other than the tivo spam shit is in progress or is about to
#start recording. If either is true, exits with a status code of 1. Supports
#as many tuners as you may have.
#
#You can adjust how far off the script should look for recordings by adding an
#argument specifying the amount of time in seconds. E.g:
#
#recstatus.tcl 300
#
#Would exit with a status code of 1 if you had an upcoming recording within 5
#minutes. Default is 600 seconds or 10 minutes. This script is a library meant
#to be run from other scripts, so there is no text output.
#

EnableTransactionHoldoff true

source /tvlib/tcl/tv/mfslib.tcl

set checkspan [lindex $argv 0]
if { $checkspan == "" } { set checkspan 600 }

set db [dbopen]

set now [clock seconds]
set onedayinsec [expr (60 * 60 * 24)]
set today [expr ($now / $onedayinsec)]
set tomorrow [expr ($today + 1)]
set timeinsectoday [expr ($now - ($today * $onedayinsec))]

#If an unwanted title is detected, return a true result
proc checkspamshit {show} {
  set prog [dbobj $show get Program]
  set title [dbobj $prog get Title]
  set sdd [string match {*TiVo Service Data*} $title]
  set sc [string match {*DVR Showcase*} $title]
  set live [string match {Live} $title]
  if { $sdd || $sc || $live } { return 1 }
  return 0
}

#If there is a recording in progress other than live or tivo spam shit, return
#a true result.
proc checkrecinprogress {} {
  global db

  ForeachMfsFile fsid name type /Recording/InProgress "" {
    RetryTransaction {
      set obj [db $db openidconstruction $fsid]
      set show [dbobj $obj get Showing]
      set showdate [dbobj $show get Date]
      if { ![checkspamshit $show] } { return 1 }
    }
  }
  return 0
}

#If an imminent recording other than the tivo spam shit is in the todo list,
#return a true result
proc checktodo {} {
  global checkspan
  global db
  global now
  global onedayinsec
  global today
  global tomorrow
  
  ForeachMfsFile fsid name type /Recording/Active "" {
    RetryTransaction {
      set obj [db $db openidconstruction $fsid]
      set show [dbobj $obj get Showing]
      set showdate [dbobj $show get Date]
      #Save time by not processing anything older than today.
      if { $showdate >= $today } {
        #Use date and time in seconds in case this script is being run at
        #midnight, that way we don't get cut off by the date rollover.
        set time [dbobj $show get Time]
        set showtime [expr (($showdate * $onedayinsec) + $time)]
        set maxrecordtime [expr ($now + $checkspan)]

        #Stop looping once we have passed the max record time
        if { $showtime > $maxrecordtime } { return 0 }

        if { $now <= $showtime && $showtime <= $maxrecordtime } {
          if { ![checkspamshit $show] } { return 1 }
        }
      }
    }
  }
  #Default
  return 0
}

#Execution starts here.

if { [checkrecinprogress] || [checktodo] } { exit 1 }

exit 0
