#!/bin/bash
#Copyright 2006 AlphaWolf_HK
#Neither this file nor any derivative may be distributed anywhere outside of
#the forums located at www.dealdatabase.com.
#
#Now we can finally reboot a tivo in a more realistic manner.
#
#This script works under the assumption that you are using the busybox ps, it
#may need modification in order to work with other versions.
#
#I would have used fuser but it doesn't allow us to selectively not kill
#certain processes due to the way tivos init system is setup.
#
#Since terminating all processes is known to crash some tivo configurations,
#this script defaults to just checking the todo list and then rebooting. If you
#want to do a full safe reboot, run the script as follows:
#
#safereboot.sh -f
#
#My personal theory is that if this script causes your tivo to crash while in
#full mode, you are running it while your init scripts are still active. They
#can run indefinitely if your init environment isn't setup properly.
#

function RomeoMustDie () {
  sync; sync; sync
  for pid in $(ps | sort -r | grep -v -w PID | grep -v -w init | grep -v bash | grep -v tnlited | grep -v safereboot | grep -v '\[.*\]' | grep -v -w sort | grep -v -w grep | grep -v -w cut | grep -v -w ps | cut -c0-5)
  do
    kill -$1 $pid 2> /dev/console
  done
  sync; sync; sync
  sleep 5
  pids=`ps | sort -r | grep -v -w PID | grep -v -w init | grep -v bash | grep -v tnlited | grep -v safereboot | grep -v '\[.*\]' | grep -v -w sort | grep -v -w grep | grep -v -w cut | grep -v -w ps | cut -c0-5`
  if [ "$pids" != "" ]; then
    return 1
  else
    return 0
  fi
}

function SanctifyTheLiving () {
  for point in $(cut -d ' ' -f2 /etc/mtab)
  do
    iteration=0
    intheloop=1
    while [ $intheloop == 1 ]
    do
      mount -o remount,ro $point 2> /dev/console
      intheloop=$?
      if [ $intheloop == 1 ]; then
        RomeoMustDie 9
      fi
      iteration=`expr $iteration + 1`
      if [ $iteration -gt 9 ]; then
        #Partition obviously doesn't want to budge, just reboot or else we'll
        #sit here forever.
        return 1
      fi
    done
  done
}

function FullReboot () {
  #Well, not really all processes, just the ones that are necessary for a clean
  #shutdown.
  echo "Sending all processes the term signal..."  
  shitleftover=0
  RomeoMustDie 15
  if [ $? == 1 ]; then
    RomeoMustDie 15
    shitleftover=$?
  fi
  if [ $shitleftover == 1 ]; then
    echo "Sending all remaining processes the kill signal..."
    RomeoMustDie 9
  fi
  echo "Remounting all filesystems as read only..."
  SanctifyTheLiving
  if [ $? == 1 ]; then
    echo "Couldn't remount filesystems; will force unclean reboot."
  else
    echo "Succeeded."
  fi
  echo "Rebooting in 10 seconds..."
}

function SignalReboot () {
  text2osd "Going down for reboot in ten seconds..."
  sleep 10
  sync; sync; sync
  reboot
}

#Execution starts here.

param="$1"

echo "Safereboot.sh by AlphaWolf_HK"
echo ""

while [ true ]
do
  echo "Checking todo list..."
  recstatus.tcl
  if [ $? == 1 ]; then
    echo "Recording in progress or about to begin, waiting to retry in 10 minutes."
    sleep 600
  elif [ "$param" == "-f" ]; then
    echo "Todo list clear, going down for full reboot:"
    #Might be safer if we give tivosh some time to clean up after itself?
    sleep 3
    FullReboot
    SignalReboot
  else
    echo "Todo list clear, rebooting in ten seconds."
    SignalReboot
  fi
done
