#!/tvbin/tivosh 
# EndPadPlus * for all your soft padding needs
# ==========================================
#
# Please see the associated README.txt for more information.
#

# The name of the file containing the PID of the backgrounded process
set pidfile "/var/run/endpad.pid"

# The name of the file with the kill instruction
set killfile "/var/run/endpad.kill"

# The name of the log file
set logfile "/var/log/endpad.log"
set Ologfile "/var/log/Oendpad.log"

# The output of the script to stderr, in case of sudden failures
set errfile "/var/log/endpad.err"

# Check if a process referenced by a PID file is running
proc checkpid pidfile {
  global errfile
  if { [file exists $pidfile] } {
    set f [open $pidfile]
    gets $f runningpid
    close $f
    if { ![file exists "/proc/$runningpid"] } {
      file delete -force $pidfile
      puts "A stale PID file was found and has been deleted."
      puts "This means that this script terminated unexpectedly last time. Loading now."
      puts "The errors logged (if any) before the last termination were:"
      if { [file exists $errfile] } {
        set f [open $errfile]
        gets $f errline
        puts " "
        while { $errline!=""} {
          puts "$errline"
          gets $f errline
        }
        puts " "
        close $f
      }
      return 1
    }
    return 0
  }
  return -1
}

# Background error catcher
proc bgerror errmessage {
  global logchan
  global errorInfo
  puts $logchan $errmessage
  puts $logchan $errorInfo
  close $logchan
}

# From mfslib.tcl with small amendment, deleting its console output
proc RetryTransaction { body } {
  global errorInfo errorCode
  
  while { 1 } {
    set code [catch {transaction {uplevel $body}} string]
    if { $code == 0 } {
      return $string
    } elseif { $code == 1 } {
      if { $errorCode == "errTmActiveLockConflict" ||
      $errorCode == "errTmBackgroundHoldoff" ||
      $errorCode == "errFsLockConflict" } {
        # puts "retrying after $errorCode ..."
        after [random 2000 4000]
        # retry the transaction
        continue 
      }
      error $string $errorInfo $errorCode 
    } elseif { $code == 2 } {
      return -code return $string
    } elseif { $code == 3 } {
      return -code break $string
    } elseif { $code == 4 } {
      return -code continue $string
    }
  }
}

# To Delete suggestions!
proc DeleteSuggestion { recfsid errstring } {
  global db
  set canceldate [expr [clock seconds] / 86400]
  set canceltime [expr [clock seconds] % 86400]

  RetryTransaction {
    set rec [db $db openid $recfsid]
    set state [dbobj $rec get State]
    if { $state != 6 } {
      return 0
    } else {
      dbobj $rec set CancelReason 10
      dbobj $rec set CancelDate $canceldate
      dbobj $rec set CancelTime $canceltime
      dbobj $rec set ErrorString $errstring
      dbobj $rec set State 2
    }
  }
  return 1
}

# determines appropriate padding on any unit
# by falcontx
# some code stolen from TivoWeb during the making of this proc
proc GetPadding { recfsid recstationfsid paddingstart paddingend padtype recseltype} {
  global db
  global num_tuners
  global version
  global seq
  global nob2b
  global verbose

  ForeachMfsFile fsid name type "/Recording/Active" "3:" {
    if { $fsid != $recfsid && (![info exists tuner1] || $num_tuners == 2 && ![info exists tuner2]) } {
      RetryTransaction {
        set crec [db $db openid $fsid]
        set cshowing [dbobj $crec get Showing]
        set cstation [dbobj $cshowing get Station]
        set cstationfsid [dbobj $cstation fsid]
        set cdate [dbobj $cshowing get Date]
        set ctime [dbobj $cshowing get Time]
        set cstartpadding [defaultval 0 [dbobj $crec get StartPadding]]
        set cendpadding [defaultval 0 [dbobj $crec get EndPadding]]
        set cduration [dbobj $cshowing get Duration]
        set cstart [expr $cdate * 86400 + $ctime]
        set cend [expr $cstart + $cduration + $cendpadding]
        if { $cend > $paddingstart && $cend < $paddingend } {
          # conflicting show ends in the middle of potential padding
          if { $verbose } { AddLog "  Active recording ends in middle of requested padding" }
          if { $padtype == "start" } {
            set ctuner [list $cend $paddingend]
          } else {
            set ctuner [list 0 0]
          }
        } elseif { $cend >= $paddingend } {
          if { $verbose } { AddLog "  Active recording lasts throughout requested padding" }
          set ctuner [list 0 0]
        } 
        if { [info exists ctuner] } {
          if { [info exists tuner1] } {
            if { $verbose } { AddLog "    Tuner 2 has [expr [lindex $ctuner 1] - [lindex $ctuner 0]] seconds available" }
            set tuner2 $ctuner
          } else {
            if { $verbose } { AddLog "    Tuner 1 has [expr [lindex $ctuner 1] - [lindex $ctuner 0]] seconds available" }
            set tuner1 $ctuner
            if { $nob2b && $num_tuners == 2 && $recstationfsid == $cstationfsid } {
              if { $verbose } { AddLog "    Active recording is on same channel; ignoring Tuner 2" }
              set tuner2 [list 0 0]
            }
          }
        }
      }
    }
  }
  set suggestions [list]
  set date [expr $paddingstart / 86400 + 1]
  ForeachMfsFile fsid name type "/Recording/Active" "4:" {
    if { $fsid != $recfsid && (![info exists tuner1] || $num_tuners == 2 && ![info exists tuner2]) } {
      scan $name "4:%d:%d:" scandate scantime
      regsub {^0+([1-9])} $scantime {\1} scantime
      set scansecs [expr $scandate * 86400 + $scantime]
      RetryTransaction {
        set nrec [db $db openid $fsid]
        if {$version >= 3} {
          set nrecbeh [dbobj $nrec get RecordingBehavior]
          set nseltype [dbobj $nrecbeh get PresentationBehavior]
        } else {
          set nseltype [dbobj $nrec get SelectionType]
        }
      }
      if { $nseltype == 6 && ($recseltype != 6 || !$seq) } {
        lappend suggestions [list $fsid $scansecs]
      } else {
        if {$scansecs <= $paddingend} {
          RetryTransaction {
            set nrec [db $db openid $fsid]
            set nshowing [dbobj $nrec get Showing]
            set nstation [dbobj $nshowing get Station]
            set nstationfsid [dbobj $nstation fsid]
            set nstartpadding [defaultval 0 [dbobj $nrec get StartPadding]]
            set nendpadding [defaultval 0 [dbobj $nrec get EndPadding]]
            set nduration [dbobj $nshowing get Duration]
            set nstart $scansecs
            set nend [expr $scansecs + $nstartpadding + $nduration + $nendpadding]
            if { [info exists ntuner] } {
              unset ntuner
            }
            if { $nstart > $paddingstart && $nstart < $paddingend } {
              # conflicting show starts in the middle of potential padding
              if { $verbose } { AddLog "  Scheduled recording starts in middle of requested padding" }
              if { $padtype == "end" } {
                set ntuner [list $paddingstart $nstart]
              } else {
                set ntuner [list 0 0]
              }
            } elseif { $nend > $paddingstart && $nend < $paddingend } {
              # conflicting show ends in the middle of potential padding
              if { $verbose } { AddLog "  Scheduled recording ends in middle of requested padding" }
              if { $padtype == "start" } {
                set ntuner [list $nend $paddingend]
              } else {
                set ntuner [list 0 0]
              }
            } elseif { $nstart <= $paddingstart && $nend >= $paddingend } {
              if { $verbose } { AddLog "  Scheduled recording lasts throughout requested padding" }
              # conflicting show lasts throughout potential padding
              set ntuner [list 0 0]
            }
            if { [info exists ntuner] } {
              if { [info exists tuner1] } {
                if { $verbose } { AddLog "    Tuner 2 has [expr [lindex $ntuner 1] - [lindex $ntuner 0]] seconds available" }
                set tuner2 $ntuner
              } else {
                if { $verbose } { AddLog "    Tuner 1 has [expr [lindex $ntuner 1] - [lindex $ntuner 0]] seconds available" }
                set tuner1 $ntuner
                if { $nob2b && $num_tuners == 2 && $recstationfsid == $nstationfsid } {
                  if { $verbose } { AddLog "    Scheduled recording is on same channel; ignoring Tuner 2" }
                  set tuner2 [list 0 0]
                }
              }
            }
          }
        }
      }
    }
  }
  
  set allowsug 0
  if { $num_tuners == 2 } {
    if { [info exists tuner2] } {
      if { [expr [lindex $tuner2 1] - [lindex $tuner2 0]] > [expr [lindex $tuner1 1] - [lindex $tuner1 0]] } {
        set ftuner $tuner2
        if { $verbose } { AddLog "  Tuner 2 has most available padding ([expr int(([lindex $ftuner 1] - [lindex $ftuner 0])/15.0)*15] seconds)" }
      } elseif { [expr [lindex $tuner1 1] - [lindex $tuner1 0]] > 0 } {
        set ftuner $tuner1
        if { $verbose } { AddLog "  Tuner 1 has most available padding ([expr int(([lindex $ftuner 1] - [lindex $ftuner 0])/15.0)*15] seconds)" }
      } else {
        if { $verbose } { AddLog "  Both tuners in use; no padding available" }
        return [list 0 [list]]
      }
    } else {
      set ftuner [list $paddingstart $paddingend]
      if { [info exists tuner1] } {
        if { $verbose } { AddLog "  Tuner 2 is available; using requested padding ([expr int(([lindex $ftuner 1] - [lindex $ftuner 0])/15.0)*15] seconds)" }
      } else {
        set allowsug 1
        if { $verbose } { AddLog "  Both tuners available; using requested padding ([expr int(([lindex $ftuner 1] - [lindex $ftuner 0])/15.0)*15] seconds)" }
      }
    }
  } elseif { [info exists tuner1] && [expr [lindex $tuner1 1] - [lindex $tuner1 0]] > 0 } {
    set ftuner $tuner1
    if { $verbose } { AddLog "  Tuner has available padding ([expr int(([lindex $ftuner 1] - [lindex $ftuner 0])/15.0)*15] seconds)" }
  } elseif { [info exists tuner1] } {
    if { $verbose } { AddLog "  Tuner in use; no padding available" }
    return [list 0 [list]]
  } else {
    set ftuner [list $paddingstart $paddingend]
    if { $verbose } { AddLog "  Tuner is available; using requested padding ([expr int(([lindex $ftuner 1] - [lindex $ftuner 0])/15.0)*15] seconds)" }
  }
    
  set paddingstart [lindex $ftuner 0]
  set paddingend [lindex $ftuner 1]
  set cancellistsug [list]

  foreach suggestion $suggestions {
    set fsid [lindex $suggestion 0]
    set scansecs [lindex $suggestion 1]
    if {$scansecs <= $paddingend} {
      RetryTransaction {
        set srec [db $db openid $fsid]
        set sshowing [dbobj $srec get Showing]
        set sstartpadding [defaultval 0 [dbobj $srec get StartPadding]]
        set sendpadding [defaultval 0 [dbobj $srec get EndPadding]]
        set sduration [dbobj $sshowing get Duration]
        set sstart $scansecs
        set send [expr $scansecs + $sstartpadding + $sduration + $sendpadding]
        if { $sstart > $paddingstart && $sstart < $paddingend ||
             $send > $paddingstart && $send < $paddingend ||
             $sstart <= $paddingstart && $send >= $paddingend } {
          lappend cancellistsug $fsid
        } 
      }
    }
  }
  set cancellistsug [lrange $cancellistsug $allowsug end]
  if { $verbose && [llength $cancellistsug] > 0} { AddLog "  Number of conflicting suggestions detected: [llength $cancellistsug]" }

  return [list [expr int(($paddingend - $paddingstart)/15.0)*15] $cancellistsug]
}

# taken from TivoWebPlus
proc defaultval {val1 val2} {
  if { $val2 != "" } {
    return $val2
  } else {
    return $val1
  }
}

# proc for adding logs, uses timezone offset
# by falcontx
proc AddLog { message } {
  global logchan
  global datestamp
  puts $logchan "[clock format [expr [set ctime [clock seconds]] + [get_tzoffset $ctime]] -format $datestamp] : $message"
}

proc FTime { time } {
  return [clock format [expr $time + [get_tzoffset $time]] -format %H:%M]
}

# proc for displaying time using timezone offset
# taken from TivoWebPlus
proc get_tzoffset { time } {
  global uktivo
  global tzoffset
  global daylightsavings
  global dstOfsList
  
  if {$daylightsavings!=2} {return $tzoffset}
  if {$time<=0} {return 0}
  incr time $tzoffset
  set idx 0
  set y [expr $time / 31536000]
  set leaps [expr ($y+2)>>2]
  set toy [expr ($time % 31536000) - $leaps*86400]
  for {set i 0} {$i<2} {incr i} {
    foreach {ddoy t} [lindex $dstOfsList $idx] break
    if {$toy < $t+86400*((73-($ddoy+$y+$leaps))%7)} break
    incr idx
  }
  #puts "$idx,$ddoy,$toy,$i=[clock format $time]"
  return [expr $tzoffset+($i&1)*3600]
}

proc get_channelnums {} {
  global guideindexdir channelnums
  RetryTransaction {
    set channeltablefsid [lindex [mfs find $guideindexdir/ChannelTable] 0]
    set channeltabledata [mfs get $channeltablefsid]
  }
  binary scan [string range $channeltabledata 0 3] I numkeys
  for {set i 0} {$i < $numkeys} {incr i} {
    binary scan [string range $channeltabledata [expr $i*8 + 4] [expr $i*8 + 11]] ISS stationfsid dummy chnum
    set channelnums($stationfsid) $chnum
  }
}


# See if we are supposed to be killing a version that is in the
# background.  If we are, create the kill file
if { [lindex $argv 0] == "-stop" } {
  if { [checkpid $pidfile]==0} {
    set pfchan [open $killfile "w"]
    puts $pfchan "Kill EndPadPlus, this file can be deleted"
    close $pfchan
    puts "Kill signal sent; EndPadPlus may take up to a minute to stop."
    puts "You will not get an acknowledgement. Please check log to confirm exit."
  } else {
    puts stderr "No instance of EndPadPlus running."
  }
  exit 0
}

set eppversion 1.2.2b

# If we get the -auto flag, may be being restarted after TiVo crash, so ignore .pid
# file and start up anyway after waiting for 2 mins for startup to complete
if { [lsearch -exact $argv "-auto"] != -1 } {
  file delete -force $pidfile
}

# If endpad is already running, then warn and stop
if {[checkpid $pidfile] == 0} {
  puts stderr "EndPadPlus is already running."
  exit 0
} 

# If we get the -run flag then this must be the forked copy
# so create the pid file and get going
# else fork a copy in the background and quit
if { [lindex $argv 0] == "-run" } {
  set pfchan [open $pidfile "w"]
  puts $pfchan [pid]
  close $pfchan
} else {
  set s [lindex $argv 0]
  set e [lindex $argv 1]
  # fork off a copy of this script to be run in the background
  if {$s == "" || $e == ""} {
    puts "EndPadPlus $eppversion"
    puts " "
    puts "Usage: /var/hack/endpadplus.tcl s e \[-seq\] \[-shd\] \[-nob2b\] \[-noverbose\] \[-auto\]"
    puts " "
    puts "Parameters:"
    puts "   s     start padding in minutes"
    puts "   e     end padding in minutes"
    puts " "
    puts "Options:"
    puts " -seq        suggestion equalization"
    puts "               (padding a suggestion will not delete another suggestion)"
    puts " -shd        record only HD suggestions"
    puts " -nob2b      back-to-back exemption"
    puts "               (second tuner will not be used to add padding to"
    puts "                 back-to-back recordings on the same channel)"
    puts " -noverbose  less detail in logs"
    puts " -auto       silent start (use when starting from a script)"
    puts " "
    exit 0
  }
  if {$s == 0 && $e == 0} {
    puts "No padding requested."
    exit 0
  }
  
  if { [lsearch -exact $argv "-auto"] == -1 } {
    puts "EndPadPlus $eppversion starting up"
    puts "Please see $logfile for details."
  }

  exec $argv0 -run $argv >& $errfile &
  exit 0
}

# The working code, only run in the forked copy

set argv [lindex $argv 1]
set startpaddingmins [lindex $argv 0]
set endpaddingmins [lindex $argv 1]
set seq [expr [lsearch -exact $argv "-seq"] != -1]
set shd [expr [lsearch -exact $argv "-shd"] != -1]
set nob2b [expr [lsearch -exact $argv "-nob2b"] != -1]
set verbose [expr [lsearch -exact $argv "-noverbose"] == -1]

set startpadding [expr round($startpaddingmins * 4)*15]
set endpadding [expr round($endpaddingmins * 4)*15]

# Open logfile
if {![file exists $logfile]} {
  set logdate [clock format [clock seconds] -format "%j"]
  set logchan [open $logfile "w"]
} else {  
  set logdate [clock format [file mtime "$logfile"] -format "%j"]
  set logchan [open $logfile "a"]
}

# Initialise stuff
set dbPoolSize [expr 100 * 1024]
set test 1
set errortime 0
set datestamp "%A %H:%M:%S"
set wakeprestart 120
set wakepreend [expr $wakeprestart + $startpadding]
catch {EnableTransactionHoldoff true}

set selectiontypes "{Show Recommendation} {Package Recommendation} {By Name} {By Channel} Manual {Suggestion} {Fuzzy Package} Bookmark {Season Pass} {Manual Season Pass} Guide IPreview WishListPass Extended ManualExtended CaptureRequest DailyData"
set presentationbehtypes "Single {Season Pass} Wishlist {Watch Only} {Tivo Clips} Suggestion {Tivo Clips Suggestion} Invisible {Manual Season Pass} Manual {Cache History}"

# get software version, number of tuners, tivo type and timezone offset
# detection code from TivoWebPlus
set db [dbopen $dbPoolSize]
RetryTransaction {
  set swsystem [db $db open /SwSystem/ACTIVE]
  set tivoswversion [dbobj $swsystem get Name]
  set setup [db $db open /Setup]
  set uktivo [expr [string index $tivoswversion [expr [string length $tivoswversion] - 1]] == 3]
  regsub -all {[[:alpha:]]} [lindex [split $tivoswversion -] 0] "" version
  if {$version >= 3} {
    set lconfig  [db $db open /State/LocationConfig]
    if { $version >=4 } {
      set setuptz [dbobj $lconfig get TimeZoneOffset]
    } else {
      set setuptz [dbobj $lconfig get TimeZoneOld]
    }
    if { $version >=5 } {
      # Assumes that UseDaylightSavings uses 0=No, 1=Yes, NA=Yes
      set daylightsavings [dbobj $lconfig get UseDaylightSavings]
      if {$daylightsavings != ""} {incr daylightsavings}
    } else {
      # DaylightSavingsPolicy uses 1=No, 2=Yes
      set daylightsavings [dbobj $lconfig get DaylightSavingsPolicy]
    }
    set AvConfig [db $db open /State/AvConfig]
  } else {
    set setup [db $db open /Setup]
    set setuptz [dbobj $setup get TimeZone]
    set daylightsavings [dbobj $setup get DaylightSavingsPolicy]
    set AvConfig [db $db open /Setup]
  }
  set num_tuners [defaultval 1 [dbobj $AvConfig get TunerCount]]
}
dbclose $db

if {$setuptz == ""} {
  set setuptz 0
}

# Determine Timezone Offset from GMT
if { $setuptz <= 0 || $version >= 4} {
  set tz $setuptz
} else {
  set tzlist "-5 -6 -7 -8 -9 -10 0 1 2 3 4 5 6 7 8 9 10 11 12 -1 -2 -3 -4 -11 -1 2"
  set tz [expr [lindex $tzlist [expr $setuptz - 1]]*60*60]
  unset tzlist
}
set tzoffset [expr round(([clock scan [lreplace [exec date -u] 4 4]] - [clock seconds] + $tz)/900.0)*900]

if {$daylightsavings == ""} {
  set daylightsavings 2
}

set dstOfsList {}
if {$uktivo} {
  #  Start: Last week in March   at 1am
  #  Ends:  Last week in October at 1am
  foreach {m w h} {4 0 1 11 0 1} {
    set doy [expr ([clock scan "$h:00 $m/1/1970" -gmt 1]/86400)+7*($w-1)]
    lappend dstOfsList [list $doy [expr $doy*86400 + $h*3600]]
  }
} else {
  #  Start: Second week in March    at 2am
  #  Ends:  First  week in November at 1am
  foreach {m w h} {3 2 2 11 1 1} {
    set doy [expr ([clock scan "$h:00 $m/1/1970" -gmt 1]/86400)+7*($w-1)]
    lappend dstOfsList [list $doy [expr $doy*86400 + $h*3600]]
  }
}

if {$version >= 3.5} {
   set guideindexdir "/GuideIndexV3"
} elseif {$version >= 3} {
   set guideindexdir "/GuideIndexV2"
} else {
   set guideindexdir "/GuideIndex"
}

AddLog " "
AddLog "EndPadPlus $eppversion starting up"
AddLog "Software version: $tivoswversion"
AddLog "Number of tuners: $num_tuners"
AddLog "Time Zone offset: [expr $tzoffset/3600] hours"
AddLog "Daylight Savings: [expr $daylightsavings == 2 ? \"on\" : \"off\"]"
AddLog "Start padding set to $startpadding seconds"
AddLog "End padding set to $endpadding seconds"
AddLog "Suggestion equalization is [expr $seq ? \"on\" : \"off\"]"
AddLog "HD suggestions only is [expr $shd ? \"on\" : \"off\"]"
AddLog "Back-to-back exemption is [expr $nob2b ? \"on\" : \"off\"]"
AddLog "Verbose logging is [expr $verbose ? \"on\" : \"off\"]"
flush $logchan

# Main loop continues forever unless kill signal recieved
while {$test} {

  # Error trap for those pesky db retry errors
  if {[catch {

    # Woken up
    AddLog " " 
    AddLog "Rise and shine!" 
    
    set db [dbopen $dbPoolSize]
    
    # reload channel data
    get_channelnums

    # Read mfs for currently recording and next schedules programmes
    RetryTransaction {
      set currecs [mfs scan "/Recording/Active" -start "3:"]
    }

    set reccount 0
    set nextepwake 0
    AddLog "Processing current recording(s):"

    #Let's see if any end padding needs doing on the current recordings
    if {$endpadding > 0} {
      foreach currecid $currecs {
        scan [lindex $currecid 1] "%d:%d:%d:%d" rectype recstartdate recstarttime recfsid
        if { $rectype == 3 } {
          incr reccount
          RetryTransaction {
            set currec [db $db openid $recfsid]
            if {$version >= 3} {
              set currecbeh [dbobj $currec get RecordingBehavior]
              set recseltype [dbobj $currecbeh get PresentationBehavior]
              set recseltypestr [lindex $presentationbehtypes [expr $recseltype - 1]]
            } else {
              set recseltype [dbobj $currec get SelectionType]
              set recseltypestr [lindex $selectiontypes [expr $recseltype - 1]]
            }
            set recshowing [dbobj $currec get Showing]  
            set recstation [dbobj $recshowing get Station]
            set recstationfsid [dbobj $recstation fsid]
            set recdate [dbobj $recshowing get Date]
            set rectime [dbobj $recshowing get Time]
            set reccallsign [dbobj $recstation get CallSign]
            set recprogram [dbobj $recshowing get Program]
            set rectitle [lindex [dbobj $recprogram get Title] 0]
            set recduration [dbobj $recshowing get Duration]
            set recendpadding [defaultval 0 [dbobj $currec get EndPadding]]
          }
          set recendtime [expr $recdate * 86400 + $rectime + $recduration + $recendpadding]
          if { [info exists channelnums($recstationfsid)] } {
            set channel "$channelnums($recstationfsid) "
          } else {
            set channel ""
          }
          AddLog "* $rectitle on \[$channel$reccallsign\] until [FTime $recendtime] ($recseltypestr)"
          if {$version < 3 || [lsearch -exact "5 7 8" $recseltype] == -1} {
            if {[clock seconds] >= [expr $recendtime - $wakepreend] } {
              if {$recendpadding < $endpadding} {
                set paddata [GetPadding $recfsid $recstationfsid $recendtime [expr $recendtime + $endpadding - $recendpadding] "end" $recseltype]
                
                set recaddpadding [lindex $paddata 0]
                set cancellistsug [lindex $paddata 1]

                # delete shows in cancellistsug; it will contain all suggestions that need to be deleted.
                if { [llength $cancellistsug] > 0 } {
                  AddLog "  Deleting suggestions: $cancellistsug"
                  foreach cancelsugid $cancellistsug {
                    DeleteSuggestion $cancelsugid "Deleting to record another program"
                  }
                  flush $logchan 
                  after 10000
                }
                # set EndPadding to recaddpadding
                if { $recaddpadding > 0 } {
                  RetryTransaction {
                    set currec [db $db openid $recfsid]
                    dbobj $currec set EndPadding [expr $recendpadding + $recaddpadding]
                  }
                  AddLog "  Adding $recaddpadding seconds end padding"
                  flush $logchan 
                  after 10000
                } else {
                  AddLog "  Can't add end padding due to conflict"
                }
              } else {
                AddLog "  Appropriate padding already exists"
              }
            } else {
              AddLog "  Not time to add end padding yet"  
              # Set time to wake up for next end padding job
              set pnextepwake [expr $recendtime - $wakepreend]
              if {[clock seconds] < $pnextepwake && ($nextepwake == 0 || $nextepwake > $pnextepwake)} {
                set nextepwake $pnextepwake
              }
            }
          } else {
            AddLog "  Will not add padding to $recseltypestr recordings"
          }
        }
      }
    } else {
      AddLog "* No end padding requested"
    }
    if { $reccount == 0 } {
      AddLog "* No recordings in progress"
    }

    RetryTransaction {
      set nextrecs [mfs scan "/Recording/Active" -start "4:" -count $num_tuners]
    }
    set nextspwake 0
    AddLog "Processing next recording(s):"

    if { $startpadding != 0 && [llength $nextrecs] > 0 } {
      foreach nextrec $nextrecs {
        scan [lindex $nextrec 1] "%d:%d:%d:%d" nextrectype nextrecstartdate nextrecstarttime nextrecfsid
        set nextstart [expr $nextrecstartdate * 86400 + $nextrecstarttime]
        RetryTransaction {  
          set nrec [db $db openid $nextrecfsid]    
          if {$version >= 3} {
            set nrecbeh [dbobj $nrec get RecordingBehavior]
            set nseltype [dbobj $nrecbeh get PresentationBehavior]
            set nseltypestr [lindex $presentationbehtypes [expr $nseltype - 1]]
          } else {
            set nseltype [dbobj $nrec get SelectionType]
            set nseltypestr [lindex $selectiontypes [expr $nseltype - 1]]
          }
          set nshowing [dbobj $nrec get Showing]
          set nstation [dbobj $nshowing get Station]
          set nstationfsid [dbobj $nstation fsid]
          set nduration [dbobj $nshowing get Duration]
          set nbits [dbobj $nshowing get Bits]
          set ncallsign [dbobj $nstation get CallSign]
          set nprogram [dbobj $nshowing get Program]
          set ntitle [lindex [dbobj $nprogram get Title] 0]
          set nstartpadding [defaultval 0 [dbobj $nrec get StartPadding]]
          set nendpadding [defaultval 0 [dbobj $nrec get EndPadding]]
        }
        set nextend [expr $nextstart + $nstartpadding + $nduration + $nendpadding]
        if {$nduration < $wakepreend} {
          set pnextepwake $nextstart
        } else {
          set pnextepwake [expr $nextstart + $nduration - $wakepreend]
        }
        if { $nextepwake == 0 || $nextepwake > $pnextepwake } {
          set nextepwake $pnextepwake
        }
        if { [info exists channelnums($nstationfsid)] } {
          set channel "$channelnums($nstationfsid) "
        } else {
          set channel ""
        }
        AddLog "* $ntitle on \[$channel$ncallsign\] from [FTime $nextstart] to [FTime $nextend] ($nseltypestr)"
        if { $shd && $nseltype == 6 && !(4096 & [expr $nbits+0]) } {
          DeleteSuggestion $nextrecfsid "Deleting standard definition suggestion"
          AddLog "  Deleting recording because it is a standard definition suggestion"
        } else {
          if { $version < 3 || [lsearch -exact "5 7 8" $nseltype] == -1 } {
            # Are we at the last possible time to add the padding 
            if {$nextstart - [clock seconds] <= [expr $startpadding + $wakeprestart]} {
              # If so then try to add start padding
              if { $nstartpadding < $startpadding } {
                set paddata [GetPadding 0 $nstationfsid [expr $nextstart + $nstartpadding - $startpadding] $nextstart "start" $nseltype]
                set recaddpadding [lindex $paddata 0]
                set cancellistsug [lindex $paddata 1]

                # delete shows in cancellistsug; it will contain all suggestions that need to be deleted.
                if { [llength $cancellistsug] > 0 } {
                  AddLog "  Deleting suggestions: $cancellistsug"
                  foreach cancelsugid $cancellistsug {
                    DeleteSuggestion $cancelsugid "Deleting to record another program"
                  }
                  flush $logchan 
                  after 10000
                }

                if {$recaddpadding > 0 } {
                  RetryTransaction {
                    # Alter the database record to add the padding
                    set nrec [db $db openid $nextrecfsid]
                    dbobj $nrec set StartPadding [expr $recaddpadding + $nstartpadding]
                    # Add Bookmark for this showing to start at the original time
                    set nbm [db $db createsub Bookmark $nrec]
                    dbobj $nbm set TimeMs [expr $recaddpadding * 1000]
                    dbobj $nrec set Bookmark $nbm                }
                  AddLog "  Adding $recaddpadding seconds start padding"
                  flush $logchan 
                  after 10000
                } else {
                  AddLog "  Can't add start padding due to conflict"
                }
              } else {
                AddLog "  Appropriate padding already exists"
              }
            } else {
              # Else set wake up for before next program starts
              AddLog "  Not time to add start padding yet"
              set pnextspwake [expr $nextstart - [expr $startpadding + $wakeprestart]]
              if {[clock seconds] < $pnextspwake && ($nextspwake == 0 || $nextspwake > $pnextspwake)} {
                set nextspwake $pnextspwake
              }
            }
          } else {
            AddLog "  Will not add padding to $nseltypestr recordings"
          }
        }
      }
    } elseif {$startpadding <= 0} {
      AddLog "* No start padding requested"
    } else {
      AddLog "* No scheduled recordings found"
      set nexttime 0
    }

    # close db until next wake up
    dbclose $db

    if { $nextepwake != 0 } {
      AddLog "Next end padding setting due [FTime $nextepwake]"
    }
    if { $nextspwake != 0 } {
      AddLog "Next start padding setting due [FTime $nextspwake]"
    }

    # Which wake up comes first?
    if { $nextspwake != 0 && $nextspwake <= $nextepwake} {
      set nexttime $nextspwake
    } else {
      set nexttime $nextepwake
    }

    # Max 30 mins between wake ups to catch new recordings
    if {$nexttime == 0 || [expr $nexttime - [clock seconds]] > 1800} {
      set nexttime [expr round([clock seconds]/60)*60 + 1800]
    }
    
    if {$nexttime < [clock seconds]} {
      set nexttime [expr [clock seconds] + 30]
    }
    
    AddLog "Next wake up will be [FTime $nexttime]"

    # switch log files at midnight
    if {$logdate != [clock format [clock seconds] -format "%j"] } {
       flush $logchan
       close $logchan
    
       catch {file copy -force $logfile $Ologfile}
       catch {file delete -force $logfile}
       set logchan [open $logfile "w"]
       AddLog "EndPadPlus log file, log rotated"
    
       set logdate [clock format [clock seconds] -format "%j"]
    }
    
    # Wait until next wake up, checking every 60s for a kill message
    AddLog "Back to sleep..."
    puts -nonewline $logchan "[clock format [expr [set ctime [clock seconds]] + [get_tzoffset $ctime]] -format $datestamp] : Checking every minute for kill signal."
    set wait 1
    while {$wait} {
      puts -nonewline $logchan "."
      flush $logchan 

      set togo [expr $nexttime - [clock seconds]]
      if {$togo <= 60} {
        after [expr $togo * 1000]
      } else {
        after 60000
      }

      if {[clock seconds] >= $nexttime} {
        set wait 0
      }

      if { [file exists $killfile] } {
        puts "EndPadPlus has stopped"
        puts $logchan " "
        AddLog "Stopping due to kill signal"
        file delete -force $pidfile
        file delete -force $killfile
        file delete -force $errfile
        flush $logchan 
        close $logchan
        exit 0
      }
    }
    
    puts $logchan " "
    flush $logchan 
  
  } result]} {
    
    # Runs if error detected
    global errorInfo guideindexdir
    AddLog "Error detected"
    puts $logchan $errorInfo
    AddLog "Attempting to recover"
    set lasterror $errortime
    set errortime [clock seconds]
    if {[expr $errortime - $lasterror] < 120} {
      # Two errors in 2 mins? Bye bye!
      AddLog "Too many errors, stopping"
      flush $logchan 
      close $logchan
      exit
    }
    # wait a bit for the db to get clear, then off we go again
    after 30000 
  }
}
