#!/tvbin/tivosh
##############################################################################
# ChrisEd BETA V1.7 Patch for folders and enhanced attribute handling
# This version for development/testing purposes only.
#
# Revision History
#  V1.1   12/01/04 First version as a plugin module/patch
#  V1.2   12/03/04  Support for station matching on SA.  Support for cross platform series matching using TmsId
#  V1.3   12/06/04  TMS Search returning objects not in MFS.  Added validation.  Added info(seriescreate) to settings.tcl
#  V1.4   12/07/04  Series Lookup was failing.  DTivo series ServerId can start with AT, AR, and ?  (was assuming AT originally)
#  V1.5   12/08/04  Found Movies with "Orphaned" series data.  The insert failed due to missing ServerId.  Corrected.
#  V1.6   09/26/05  Use existing program when available, assign program server id.
#  V1.6.1 09/26/05  Added TMS search for station (due to varying serverversions on systems)
#  V1.6.2 11/21/05  Missed initializing ProgramServerID/ProgramServerVersion (caused crosslinked shows when programserverid missing and previous show had it)
#  V1.7   01/19/06  GuideIndex dynamic by Tivo version
#
##############################################################################

proc parse_xml { xmldata } {
global info ; set p 3
outd 2 "parse_xml:"
 set info(attributes) "Title Description EpisodeTitle EpisodeNum AltEpisodeNum StartDate StopDate ExpirationDate StartTime StopTime ExpirationTime Duration Date OriginalAirDate CallSign Name Time TvRating MpaaRating Bits IsEpisode MovieRunTime MovieYear ShowType ThumbData Premiere"
 set info(xattributes) "ProgramServerId SeriesServerId StationServerId ProgramServerVersion SeriesServerVersion StationServerVersion SeriesTmsId ProgramTmsId StationTmsId Actor Advisory Director Genre Writer Category SeriesGenre GuestStar Host ExecProducer Producer"
 foreach item $info(attributes) { set info($item) "" }
 foreach item $info(xattributes) { set info($item) "" }
 set x_tags "Actor Advisory Director Writer Category GuestStar Host ExecProducer Producer" 
 set cso_tags "Begin CommercialSkipOffset End" ; set info(parts_csos) ""
 set xml_lines [split $xmldata "\n"]
 set curobjecttype "" ; set prevobjecttype ""

 foreach line $xml_lines {
  foreach item $info(attributes) {if {[regexp "<$item>" $line]} {set info($item) [item_from_xml_line $line $item]}}
  foreach item $cso_tags {
   if {[regexp "<$item>" $line]} {
	set data [item_from_xml_line $line $item]
	switch $item {
	 Begin	{ append info(parts_csos) "\{$data " }
	 End	{ append info(parts_csos) "$data\} " }
	 CommercialSkipOffset { append info(parts_csos) "$data " }
	}
   }
  }
       
#  ServerId (and otheres) are nasty to parse because they're not uniquely named.  This process may fail under deeper nesting conditions.  
   if {[regexp "<Object type=" $line]} { set prevobjecttype "$curobjecttype" ; set curobjecttype [object_from_xml_line $line] }
   if {[regexp "</Object>" $line]}     { set curobjecttype "$prevobjecttype" ; set prevobjecttype "" }
   if {[regexp "<ServerId>" $line]} {
        switch $curobjecttype {
          Series  { set info(SeriesServerId) [item_from_xml_line $line ServerId] }
          Program { set info(ProgramServerId) [item_from_xml_line $line ServerId] }
          Station { set info(StationServerId) [item_from_xml_line $line ServerId] }
        } 
   }
   if {[regexp "<ServerVersion>" $line]} {
        switch $curobjecttype {
          Series  { set info(SeriesServerVersion) [item_from_xml_line $line ServerVersion] }
          Program { set info(ProgramServerVersion) [item_from_xml_line $line ServerVersion] }
          Station { set info(StationServerVersion) [item_from_xml_line $line ServerVersion] }
        } 
   }
   if {[regexp "<TmsId>" $line]} {
        switch $curobjecttype {
          Series  { set info(SeriesTmsId) [item_from_xml_line $line TmsId] }
          Program { set info(ProgramTmsId) [item_from_xml_line $line TmsId] }
          Station { set info(StationTmsId) [item_from_xml_line $line TmsId] }
        } 
   }
   if {[regexp "<Genre>" $line] && $curobjecttype == "Program"} { 
     set data [item_from_xml_line $line Genre]
     if {$info(Genre) == ""} {set info(Genre) $data } else {append info(Genre) ":$data" }
   }
   if {[regexp "<Genre>" $line] && $curobjecttype == "Series"} { 
     set data [item_from_xml_line $line Genre]
     if {$info(SeriesGenre) == ""} {set info(SeriesGenre) $data } else {append info(SeriesGenre) ":$data" }
   }
   foreach item $x_tags {
    if {[regexp "<$item>" $line]} {
      set data [item_from_xml_line $line $item]
      if {$info($item) == ""} {set info($item) $data } else {append info($item) ":$data" }
      }
    }
       
 }
 foreach item $info(attributes) { outd $p "$item is \"$info($item)\"" }
 foreach item $info(xattributes) { outd $p "$item is \"$info($item)\"" }
 foreach part $info(parts_csos) { outd $p "\"$part\"" }
}

proc object_from_xml_line { line } {
 set il [expr [string length "<Object type="] + 1]
 set sindex  [expr [string first "<Object type=" $line] + $il]
 set sindex2 [expr [string first " id=" $line] - 2]
 set data [string range $line $sindex $sindex2]
 return $data
}

proc findseries { tmsid serverid serverversion } {
global db

#example DTivo path  /Server/ATSH225421:4:0:0
set seriespath "/Server/$serverid:4:$serverversion:0"
outd 3 "Try DTivo series path: $seriespath"

#find and validate tmsid using fast approach 
if {[catch {RetryTransaction {
  set seriesfsid [lindex [mfs find $seriespath] 0]
  } } 1] != 0 } { set seriesfsid "" }

if { $seriesfsid != "" } {
  outd 3 "series found by ServerId = $seriesfsid"
  return $seriesfsid
  }

#example SA2 path    /Server/17028
set seriespath "/Server/$serverid"
outd 3 "Try SA series path: $seriespath"

#find and validate tmsid using fast approach 
if {[catch {RetryTransaction {
  set seriesfsid [lindex [mfs find $seriespath] 0]
  } } 1] != 0 } { set seriesfsid "" }

if { $seriesfsid != "" } {
  outd 3 "series found by ServerId = $seriesfsid"
  return $seriesfsid
  }


# find using TmsId
set seriesfsid [tms_to_fsid $tmsid]
outd 3 "series found by Tms Search = $seriesfsid"

if { $seriesfsid == "" } { return $seriesfsid }

#Validate object (apparently can be in Tms Search, but not in MFS!! )
if {[catch {RetryTransaction {
  set tempseries [db $db openid $seriesfsid]
  } } 1] != 0 } { 
  outd 3 "series missing in MFS.  discarding TMS Search result"
  set seriesfsid "" 
  }

return $seriesfsid

}

proc findstation { tmsid } {
global db

set maxversion ""
set maxfsid ""
set maxpath ""


ForeachMfsFile fsid name type "/StationTms" "$tmsid:" {

RetryTransaction {
  set obj [db $db openidconstruction $fsid]
  set version [dbobj $obj get ServerVersion]
  set path [lindex [dbobj $obj get IndexPath] 0]
  if {$version > $maxversion} {
    set maxversion $version 
    set maxfsid $fsid 
    set maxpath $path
	}
} 
}

return $maxpath
}

proc setrecinfo { recfsid } {
global info db ; set p 1
 outd $p "setrecinfo: recobj\{$recfsid\}"
 set progitems "Title Bits Description EpisodeTitle EpisodeNum AltEpisodeNum IsEpisode MovieRunTime MovieYear MpaaRating OriginalAirDate ShowType TvRating"
 set showingitems "Date Duration Time Premiere"
 set recitems "StartDate StartTime StopDate StopTime"
 set stationitems "CallSign Name"

#Is program already on system?  Yes, use it.  No, then create it with serverid
set programfsid ""
#example DTivo path  /Server/A002a3a6f:3:1:0
if {$info(ProgramServerId) != ""} {
  set programpath "/Server/$info(ProgramServerId):3:$info(ProgramServerVersion):0"
  outd 3 "Try DTivo program path: $programpath"

  #find and validate tmsid using fast approach 
  if {[catch {RetryTransaction {
    set programfsid [lindex [mfs find $programpath] 0]
    } } 1] != 0 } { set programfsid "" }
  if {$programfsid != ""} {outd 3 "Existing program path found" }
  }


#Validate station and series information before main transaction to avoid transaction conflicts
set seriesfsid ""
if {$info(SeriesTmsId) != ""} {
  set seriesfsid [findseries $info(SeriesTmsId) $info(SeriesServerId) $info(SeriesServerVersion) ]
  }

set stationpath ""
if {$info(StationServerId) != ""} {
  if {[string match "AStation*" $info(StationServerId)]} {
    #example DTivo path  /Server/AStation11097:5:61:0
    set stationpath "/Server/$info(StationServerId):5:$info(StationServerVersion):0"
    } else {
    #example SA path  /Server/380
    set stationpath "/Server/$info(StationServerId)"
    }

  outd 3 "station path: $stationpath"
  if {[catch {RetryTransaction {mfs find "$stationpath"}} l] != 0} {
     outd 3 "station path not found.  Leaving artificial station"
     set stationpath ""
     }

#try close match (DTivo)
  if {$stationpath == ""} {
    set stationpath [findstation $info(StationTmsId)]
    outd 3 "close station path found and used.$stationpath"
    }
  }



 RetryTransaction {
set rec [db $db openid $recfsid] ; set showing [dbobj $rec get Showing]
set program [dbobj $showing get Program] ; set station [dbobj $showing get Station]

if {$stationpath != ""} {
  set oldstation $station
  set stationindex [dbobj $oldstation get IndexPath]
  outd 3 "temp station discarded $stationindex"
  set station [db $db open $stationpath]
  dbobj $showing set Station $station 
  dbobj $oldstation markasrubbish
} else {
  #dbobj $station set TmsId $info(StationTmsId)  #This was bad.  Got indexed in real station area. might have side effects.
  }

foreach item $showingitems { if { $info($item) != "" } { dbobj $showing set $item $info($item) } }
foreach item $recitems { if { $info($item) != "" } { dbobj $rec set $item $info($item) } }
foreach item $stationitems { if { $info($item) != "" } { dbobj $station set $item $info($item) } }

if {$programfsid != ""} {
  set oldprogram $program
  set programindex [dbobj $oldprogram get IndexPath]
  outd 3 "temp program discarded $programindex"
  set program [db $db openid $programfsid]
  dbobj $showing set Program $program 
  dbobj $oldprogram markasrubbish
} else {

#make new program permanent when possible
if {$info(ProgramServerId) != ""} {
  dbobj $program set ServerId $info(ProgramServerId)
  dbobj $program set ServerVersion $info(ProgramServerVersion)
  }

foreach item $progitems { if { $info($item) != "" } { dbobj $program set $item $info($item) } }
dbobj $program set TmsId $info(ProgramTmsId)

foreach item "Actor Advisory Director Genre Writer GuestStar Host ExecProducer Producer" {
  set x_lines [split $info($item) ":"]
  foreach itemv $x_lines { 
    outd 3 "$item :: $itemv"
    dbobj $program add $item $itemv 
    }
  }

if {$info(Category) != ""} {
  set apgprogram [db $db createsub ApgProgram $program]
  set x_lines [split $info(Category) ":"]
  foreach itemv $x_lines { 
    outd 3 "Category :: $itemv"
    dbobj $apgprogram add Category $itemv 
    }
  dbobj $program set ApgProgram $apgprogram
  }

if {$seriesfsid != ""} {
  set series [db $db openid $seriesfsid]
  set foundtmsid [dbobj $series get TmsId]
  outd 3 "series tms check: $info(SeriesTmsId) $foundtmsid "
  if { $info(SeriesTmsId) == $foundtmsid } { 
    dbobj $program set Series $series 
    } else { set seriesfsid "" }
  }

#Create missing series only when Series TmsId and ServerId are available.
if {$seriesfsid == "" && $info(SeriesTmsId) != "" && $info(SeriesServerId) != "" && $info(seriescreate) } {
    outd 3 "Series $info(SeriesTmsId) : $info(SeriesServerId) Created"
    set series [db $db create Series]
    dbobj $program set Series $series
  #  dbobj $series set Version ??  #Missing in xml.  Let tivo assign one (hopefully)
    dbobj $series set ServerId $info(SeriesServerId)
    dbobj $series set ServerVersion $info(SeriesServerVersion)
    dbobj $series set ThumbData $info(ThumbData)
    dbobj $series set Title $info(Title)
    dbobj $series set TmsId $info(SeriesTmsId)
  
    set x_lines [split $info(SeriesGenre) ":"]
    foreach itemv $x_lines { 
      outd 3 "SeriesGenre :: $itemv"
      dbobj $series add Genre $itemv 
      }
    }

} 
#End of Else from use existing program

# End of RetryTransaction  
 }
outd $p "\{$info(Title)\}\{$info(EpisodeTitle)\}\n$info(Description)"
}

proc make_blank_rec { ProgramTitle } {
global info version db ; set p 1
outd $p "make_blank_rec: \"$ProgramTitle\""
 if { $ProgramTitle == "" } { set ProgramTitle "unknown title - inserting" }
 set date [expr [clock seconds] / 86400]
 set time [expr [expr [clock seconds] % 86400] ]
 RetryTransaction {
  set recording [db $db create Recording]
  set recordingfsid [ dbobj $recording fsid ]
  dbobj $recording set BitRate 0
  dbobj $recording set ErrorString "mfs_ftp insert"
  dbobj $recording set ExpirationTime 0 ; 
  dbobj $recording set RecordQuality 100
  dbobj $recording set Score 100 ; dbobj $recording set State 3
  if { $info(saveuntil) == "delete" } {
   dbobj $recording set ExpirationDate 24855
   dbobj $recording set SelectionType 4
  } else {
   dbobj $recording set ExpirationDate [expr $date + 7]
   dbobj $recording set SelectionType 6
  }
  dbobj $recording set StartDate $date ; dbobj $recording set StopDate $date
  dbobj $recording set StartTime $time
  dbobj $recording set StopTime [expr $time + 60]
  set station [db $db create Station]
  dbobj $station set CallSign	"mfs_ftp"
  dbobj $station set Name	"mfs_ftp"
  if { $version >= 3 } {
	dbobj $recording set UsedBy 1
	set recordingbehavior [db $db createsub RecordingBehavior $recording]
	foreach Behavior {DiskBehavior PresentationBehavior ProgramGuideBehavior} { dbobj $recordingbehavior set $Behavior 1 }
	dbobj $recordingbehavior set TunerBehavior 3
	dbobj $recording set RecordingBehavior $recordingbehavior
  }
  set showing [db $db createsub Showing $recording]
#  dbobj $recording set ActualShowing $Showing
  dbobj $showing set Date $date ; dbobj $showing set Duration 60
  dbobj $showing set Station $station ; dbobj $showing set Time $time
  dbobj $recording set Showing $showing ; dbobj $recording set StreamFileSize 0
  set program [db $db create Program]
  dbobj $program set Title "$ProgramTitle" ; dbobj $program set OriginalAirDate "$date"
  dbobj $showing set Program $program
 }
 outd $p "created blank recording \{$recordingfsid\}" ; return $recordingfsid
}



##############################################################################
# The following code was borrowed from TivoWebPlus.  index.itcl  
# Bugs were corrected in tms_to_fsid
##############################################################################

proc get_recordindexsearch {indexfsid startoffset maxoffset recsize incrsize searchkey searchtype matchsize} {
  set match 0
  for {set offset $startoffset} {$offset < $maxoffset} {incr offset $incrsize} {
    set end $incrsize
    if { $maxoffset - $offset < $incrsize } {
      set end [expr $maxoffset - $offset]
    }
    RetryTransaction {
      set indexdata [mfs getpart $indexfsid $offset $end]
    }
    set slen [expr $end / $recsize]
    set top $slen
    set current [expr $top / 2]
    set bottom 0
    set count 0
    while { $top >= $bottom } {
      set data [string range $indexdata [expr $current * $recsize] [expr $current * $recsize + $matchsize]]
      #binary scan $data H8 fsidnum
      #puts "$searchtype $offset $top $bottom $current [expr 0x$fsidnum]"
      set compare [string compare $data $searchkey]
      if {$compare == 0 || ($searchtype == 0 && $compare == 1)} {
        if { $current <= 0 } {
          set match 1
          break
        }
        set data2 [string range $indexdata [expr ($current-1) * $recsize] [expr ($current-1) * $recsize + $matchsize]]
        if {[string compare $data2 $searchkey] == -1} {
          set match 1
          break
        } else {
          set top [expr $current - 1]
        }
      } elseif {$compare == -1} {
        set bottom [expr $current + 1]
      } else {
        set top [expr $current - 1]
      }
      set current [expr $bottom + ($top - $bottom) / 2]
      incr count
      if { $count > $slen } {
        error "Error: infinite loop in binary search ($bottom $top $current $slen)"
      }
    }
    if { $match == 1} {
      if {$searchtype == 2} {
        set data [string range $indexdata [expr $current * $recsize + $matchsize + 1] [expr ($current+1) * $recsize - 1]]
        return $data
      } else {
        return [expr ($offset-$startoffset)/$recsize + $current]
      }
    }
  }

  if {$searchtype == 2} {
    return ""
  } else {
    return -1
  }
}

proc tms_to_fsid {tmsid} {
#  global guideindexdir
#  global db
  global version
  if {$version >= 6} { set guideindexdir "/GuideIndexV3" } else { set guideindexdir "/GuideIndexV2" }

  RetryTransaction {
    set tmskeyfsid [lindex [mfs find $guideindexdir/Tms.key] 0]
    set tmsindex [lindex [mfs find $guideindexdir/Tms.index] 0]
    set tmskeydata [mfs getpart $tmskeyfsid 0x0 0x18]
  }
  
  binary scan [string range $tmskeydata 0x00 0x03] I recsize
  binary scan [string range $tmskeydata 0x04 0x07] I totalnumkeys
  binary scan [string range $tmskeydata 0x08 0x0B] I incrsize 
  binary scan [string range $tmskeydata 0x10 0x13] I numkeys 

  set prefix [string range $tmsid 0 1]
  set number [string range $tmsid 2 end]

  if [expr [string length $number] == 10] {
    set mn 0
    if {[string compare "4294967296" $number] <= 0} {
      set numbers [bignum_divide $number "4294967296"]
      set n1 [lindex $numbers 0]
      set n2 [lindex $numbers 1]
    } else {
    set mn 256
    set n1 0
    set n2 $number
    }   
  } else {
    set mn 256
    set n1 0
    set n2 $number
  }

  regsub ^0+(.+) $n1 \\1 n1
  regsub ^0+(.+) $n2 \\1 n2

  set search [binary format a2SII $prefix $mn $n1 $n2]

  # look through the key-file to get a starting point
  set startoffset 24
  set maxoffset [expr $startoffset + $recsize * $numkeys]
  set first [get_recordindexsearch $tmskeyfsid $startoffset $maxoffset $recsize $incrsize $search 0 11]
  if { $first == -1 } {
    error "Error couldn't find starting offset in tms key"
  }

  # next, go throught index-file to get the exact match
  set startoffset [expr $first * $incrsize]
  set maxoffset [expr $startoffset + $incrsize]
  if { $maxoffset > $recsize * $totalnumkeys } {
    set maxoffset [expr $recsize * $totalnumkeys]
  }
  set fdata [get_recordindexsearch $tmsindex $startoffset $maxoffset $recsize $incrsize $search 2 11]
  if {[string length $fdata] == 0} {
    return ""
  }
  binary scan $fdata I fsid
  return $fsid
}

proc bignum_subtract {a b} {
  set la [string length $a]
  set lb [string length $b]

  if {$la < $lb} {
    set formatstring [format "%%0%s%s" $lb "s"]
    set a [format $formatstring $a]
    set la $lb
  } else {
    set formatstring [format "%%0%s%s" $la "s"]
    set b [format $formatstring $b]
    set lb $la
  }

  set result ""
  set carry 0
  for {set index [expr $lb-1]} {$index >= 0} {incr index -1} {
    set current [expr [string index $a $index] - [string index $b $index]]
    if {$carry == 1} {
      incr current -1
    }
    if {$current < 0} {
      set carry 1
      incr current 10
    } else {
      set carry 0
    }

    set result [format "%s%s" $current $result]

  }
  if {$carry == 1} {
    set result [format "-%s" $result]
  }
  return $result
}

# calculate dividend / divisor
proc bignum_divide {dividend divisor} {
  set result 0
  set temp $dividend

  while {[string index $temp 0] != "-"} {
    incr result 1
    set prevtemp $temp
    set temp [bignum_subtract $temp $divisor]
  }
  return "[expr $result - 1] $prevtemp"
}