#!/bin/bash
# OzTiVo S3 installer
# Steps:
# Id disk
# patch kernel
# Add hacks:
#   /hack/bin
#   ticketmaster/router+/tcpdump
#   Copy packages for mfsftp, tivoweb and bufferhack
#   telnet,ssh, bash on serial
#   rc scripts
#   patch tivoapp
#   modify bootpage 
#   set tclient.conf
# Version History
# 0.3 - First working version
# 0.4 - Added ssh password, ssh + scp clients. Dialcode to tclient.conf, 30 second skip
# 0.5 - Added bootpage option to disable software upgrade. 
#       Added tcpdump, dbset, dbget 
#       Copy mfsftp, tivoweb and bufferhack to /hack
#       Added swupdate.sh
# 0.6 - Added support for 11.3b10
#       Backup MFS keyring at first boot
#       Disable ticketmaster for 11.3b10
#       Allow update over previosu install
#       Add tclient.conf for devhd.oztivo.net (150)
#       Call /hack/etc/rc.extra at end of init scripts
#       Allow blank ssh password
VERSION=0.6

HDPARM=/sbin/hdparm
INSTALLER_DIR=$(realpath $(dirname $0))
PACKAGE_DIR=${INSTALLER_DIR}/packages
TIVOPART=$INSTALLER_DIR/bin/tivopart.x86
BOOTPAGE=$INSTALLER_DIR/bin/bootpage
REPLACE_INITRD=$INSTALLER_DIR/bin/replace_initrd
NULL_INITRD=$INSTALLER_DIR/packages/null-linuxrc.img.gz
HACKS=$INSTALLER_DIR/packages/hacks.tgz
BIN_FILES="tcpdump swupdate.sh dbset dbget mount_usb backup_keys.sh station_debug.tcl"
ETC_FILES="oztivo.config OzTiVo-SW-Update.pub"
EXTRA_PACKAGES="BufferHack.zip endpadplus-1.2.2b.zip mfsftp.tar.gz"

KERNEL_BAK=${INSTALLER_DIR}/backup/tivo_kernel.bak
MOUNTPOINT=/tmp/tivo_mount

TIVOAPP=$MOUNTPOINT/tvbin/tivoapp

# Dialog height and width
H=0
W=0
TITLE="TiVo-S3-Installer-$VERSION"
DIALOG="dialog --backtitle ${TITLE}"

#LOGFILE=/tmp/tivo_install_$(date +%s).log
LOGFILE=${INSTALLER_DIR}/logs/tivo_install.log

log() {
	local d=$(dirname $LOGFILE)
	test -f $d || mkdir -p $d
	echo "$(date) $@" >> $LOGFILE
}

info() {
	log $@
	$DIALOG --title "Info" --infobox "$@" $H $W
}

msg() {
	log $@
	$DIALOG --title "Message" --msgbox "$1" $H $W
}

cleanup() {
	cd $INSTALLER_DIR
	mount | grep -q $MOUNTPOINT 
	test  $? -eq 0   && umount $MOUNTPOINT
	test -d $MOUNTPOINT && rmdir $MOUNTPOINT
	log "END"
	while true; do
		choice=$($DIALOG --no-cancel --title "What do you want to do next?" --menu "Please choose from the following selections" $H $W 3 1 "Show log file" 2 "Shutdown" 3 "Quit to Shell" 3>&1 1>&2 2>&3) 
		case $choice in
			1) $DIALOG --title "Log file" --textbox $LOGFILE $H $W ;;
			2) poweroff ;;
			3) exit ;;
		esac
	done
}

do_exit() {
	log "$@"
	cleanup
	exit 0
}

fail() {
	$DIALOG --title "An error has occured" --msgbox "$@" $H $W
	do_exit $@
}

welcome() {
	$DIALOG --title "Welcome to the OzTivo S3 installer" --yesno "This script will install the hacks required to enable your S3 TiVo to work after October 2017\n\nDo you want to continue?" $H $W
	return $?
}

select_drive() {
	unset items
	declare -a items
	count=0
	i=0
	for d in /dev/sd?
	do
		items[i]="$d"
		((i++))
		items[i]="$(fdisk -l $d | grep "^Disk $d")"
		((i++))
		((count++))
	done
	DISK=$($DIALOG --notags --title "TiVo disk selection" --menu "Select the TiVo Drive from the list below" $H $W $count "${items[@]}" 3>&1 1>&2 2>&3) 
	return $?
}

check_prereq() {
	test -x $TIVOPART || fail "Missing tivopart tool"
	test -x $BOOTPAGE || fail "Missing bootpage tool"
	test -x $REPLACE_INITRD || fail "Missing replace_initrd tool"
	test -f $NULL_INITRD || fail "Missing initrd package"
}

check_root() {
	if [ $(id -u) -ne 0 ]; then
		fail "This script must be run as root" 
	fi
	return 0
}

get_password() {
	unset TIVO_ROOT_PASSWORD
	while true; do
		local pw1=$($DIALOG --no-cancel --title "Enter password" --passwordbox "Enter a new password for root user on TiVo (password will not be displayed)" $H $W 3>&1 1>&2 2>&3) 
		local pw2=$($DIALOG --no-cancel --title "Verify password" --passwordbox "Enter the same password again to verify (password will not be displayed)" $H $W 3>&1 1>&2 2>&3) 
		if [ "x$pw1" != "x$pw2" ]; then
			msg "Passwords do not match"
		elif [ "x$pw1" == "x" ]; then
			msg "Password is blank"
			return 1
		else
			break
		fi
	done
	TIVO_ROOT_PASSWORD=$(openssl passwd -1 ${pw1})
	return 0
}

check_drive() {
	TIVO_DRIVE=""
	log "Checking $DISK using $TIVOPART"
	info=$($TIVOPART d $DISK 2>&1)
	ret=$?
	echo $info >> $LOGFILE
	if [ $ret -ne 0 ]; then
		msg  "Non TiVo Disk\n\n$DISK does not look like a TiVo disk" 
		return 0
	fi
	$DIALOG --scrolltext --title "TiVo Disk Found" --yesno "This looks like a TiVo disk.\nDo you want to use this disk?\n\n$info" $H $W
	if [ $? -eq 0 ]; then
		TIVO_DRIVE=$DISK
	fi
	return 0
	
}

confirm_drive() {
	$DIALOG --defaultno --title "!***Warning***!" --yesno "You are about to write data to $DISK\n\nThere is a chance that this will make your disk unusable. If you have multiple disks in your PC, and you select the wrong one, then you will loose data!\n\nDo you still want to continue?" $H $W
	return $?
}

patch_kernel() {
	kernpart=$TIVO_DRIVE$($BOOTPAGE -b $TIVO_DRIVE)
	bootparams=$($BOOTPAGE -p $TIVO_DRIVE)
	info "Active kernel on $TIVO_DRIVE is in partition $kernpart\nBoot parameters are $bootparams\nPatching kernel"
	$TIVOPART r $TIVO_DRIVE 2>&1 >> $LOGFILE || fail "Could not read TiVo partition map"
	test -b $kernpart || fail "Block device $kernpart does not exist"
	result=$($REPLACE_INITRD $kernpart $NULL_INITRD $KERNEL_BAK 2>&1)
	ret=$?
	echo $result >> $LOGFILE 
	test $ret -eq 0 || fail "Could not patch kernel in $kernpart"
}

install_hacks() {
	info "Installing hacks"
	rootpart=$TIVO_DRIVE$(($($BOOTPAGE -b $TIVO_DRIVE)+1))
	info "Mounting TiVo root filesystem on $MOUNTPOINT" 
	mkdir -p $MOUNTPOINT || fail "Could not create mount point $MOUNTPOINT"
	mount $rootpart $MOUNTPOINT  || fail "Could not mount $rootpart on $MOUNTPOINT"
	test -f $TIVOAPP || fail "This does not look like a TiVo root filesystem (missing $TIVOAPP)"
	test -d $MOUNTPOINT/hack || mkdir $MOUNTPOINT/hack 
	test -d $MOUNTPOINT/hack || fail "Can't create $MOUNTPOINT/hack"

	test -e ${TIVOAPP} || fail "Could not find tivoapp on the drive"
	info "Checking tivoapp"
	PATCH_MSG="Installed tivoapp patches:"
	tmenable=""
	cd $(dirname ${TIVOAPP})
	if [  -f ${TIVOAPP}.patched ]; then
		msg "Patched tivoapp already found. Not re-patching"
		PATCH_MSG="$PATCH_MSG None"
	elif echo "668262e03665c3e1a4d4b2c0fa7018c1  tivoapp" | md5sum -c  >> $LOGFILE 2>&1
	then
		info "Patching tivoapp 11.3b8-01-2"
		cp $TIVOAPP ${TIVOAPP}.patched
		mv $TIVOAPP ${TIVOAPP}.orig
		#info "NoCSO"
		#echo -ne "\x10\x00\x00\xa2" | dd conv=notrunc of=${TIVOAPP}.patched bs=1 seek=9509444
		info "backdoors"
		echo -ne "\x24\x10\x00\x01" | dd conv=notrunc of=${TIVOAPP}.patched bs=1 seek=3381380
		PATCH_MSG="$PATCH_MSG\n - backdoors"
		#info "CCI"
		#echo -ne "\x00\x00\x80\x21" | dd conv=notrunc of=${TIVOAPP}.patched bs=1 seek=2239876
		#echo -ne "\x24\x11\x00\x00" | dd conv=notrunc of=${TIVOAPP}.patched bs=1 seek=2239912
		#echo -ne "\x00\x00\x80\x21" | dd conv=notrunc of=${TIVOAPP}.patched bs=1 seek=12088628
		#info "Delete this Recording?"
		#echo -ne "\x10\x00\x00\x03" | dd conv=notrunc of=${TIVOAPP}.patched bs=1 seek=5611984
		info "30-sec Skip"
		echo -ne "\x10\x40\x00\x26" | dd conv=notrunc of=${TIVOAPP}.patched bs=1 seek=7277108
		PATCH_MSG="$PATCH_MSG\n - 30 Second Skip"
		#info "No Thanks"
		#echo -ne "\x10\x00\x00\x08" | dd conv=notrunc of=${TIVOAPP}.patched bs=1 seek=3686148
		info "Linking to patched app"
		ln -s $(basename ${TIVOAPP}).patched $(basename ${TIVOAPP})
		ls -l ${TIVOAPP}* >> $LOGFILE
		tmenable="ticketmaster"
	elif echo "0859155bf05599ee30feeccd51205eb0  tivoapp" | md5sum -c  >> $LOGFILE 2>&1
	then
		info "Patching tivoapp 11.3b10.2017.10.11-1506-01-2-Z"
		cp $TIVOAPP ${TIVOAPP}.patched
		mv $TIVOAPP ${TIVOAPP}.orig
		info "backdoors"
		echo -ne "\x24\x10\x00\x01" | dd conv=notrunc of=${TIVOAPP}.patched bs=1 seek=3381656
		PATCH_MSG="$PATCH_MSG\n - backdoors"
		info "30-sec Skip"
		echo -ne "\x10\x40\x00\x26" | dd conv=notrunc of=${TIVOAPP}.patched bs=1 seek=7310200
		PATCH_MSG="$PATCH_MSG\n - 30 Second Skip"
		info "Linking to patched app"
		ln -s $(basename ${TIVOAPP}).patched $(basename ${TIVOAPP})
		ls -l ${TIVOAPP}* >> $LOGFILE
	else
		msg "Unrecognised tivoapp version. Not re-patching"
		PATCH_MSG="$PATCH_MSG None"
	fi

	info "Installing hack/bin"
	cd $MOUNTPOINT/hack
	tar xf $PACKAGE_DIR/hack_bin.tar.gz
	info "Creating /etc/passwd for ssh server"
	get_password
	if [ $? -eq 0 ] ; then
		info "Root password was set"
		echo "root:${TIVO_ROOT_PASSWORD}:0:0:root:/:/bin/bash" > $MOUNTPOINT/etc/passwd
	else
		info "Root password is blank"
		echo "root::0:0:root:/:/bin/bash" > $MOUNTPOINT/etc/passwd
	fi
	info "Creating /etc/shells for ssh server"
	echo "/bin/bash" > $MOUNTPOINT/etc/shells
	info "Creating /.ssh for ssh server"
	mkdir $MOUNTPOINT/.ssh
	chmod 0600 $MOUNTPOINT/.ssh
	info "Installing ticketmaster" 
	cp $PACKAGE_DIR/routerplus-2.4.20.o  $MOUNTPOINT/lib/modules
	mv $MOUNTPOINT/lib/modules/router.o $MOUNTPOINT/lib/modules/router_orig.o
	cd $MOUNTPOINT/lib/modules
	ln -s routerplus-2.4.20.o router.o
	cp $PACKAGE_DIR/ticketmaster  $MOUNTPOINT/hack/bin
	for file in ${EXTRA_PACKAGES}; do
		info "Copying $file"
		cp $PACKAGE_DIR/$file  $MOUNTPOINT/hack/
	done
	info "Installing rc scripts"
	cp $PACKAGE_DIR/rc.Sequence_255.Hacks.sh  $MOUNTPOINT/etc/rc.d/StageD_PreMfs
	chmod +x $MOUNTPOINT/etc/rc.d/StageD_PreMfs/rc.Sequence_255.Hacks.sh 
	cp $PACKAGE_DIR/rc.sysinit.author $MOUNTPOINT/etc/rc.d
	chmod +x $MOUNTPOINT/etc/rc.d/rc.sysinit.author
	info "Installing bash files" 
	cp $PACKAGE_DIR/.bash_logout $MOUNTPOINT
	cp $PACKAGE_DIR/.profile $MOUNTPOINT
	info "Installing etc files" 
	mkdir -p $MOUNTPOINT/hack/etc
	for file in $ETC_FILES; do
		info "Copying $file"
		cp $PACKAGE_DIR/$file $MOUNTPOINT/hack/etc
	done
	info "Installing bin files" 
	for file in $BIN_FILES; do
		info "Copying $file"
		cp $PACKAGE_DIR/$file $MOUNTPOINT/hack/bin
	done
	info "Setting crontab" 
	mkdir -p $MOUNTPOINT/hack/etc/crontabs
	cat > $MOUNTPOINT/hack/etc/crontabs/root <<-EOF
	5 15 * * * /hack/bin/swupdate.sh -r > /var/log/swupdate.log 2>&1
	EOF
	info "Creating USB mount point" 
	mkdir -p $MOUNTPOINT/mnt/usb
	grep "^140::" $MOUNTPOINT/etc/tclient.conf >> ${LOGFILE}
	if [ $? -ne 0 ]; then
		info "Adding DailConfig 140 to tclient.conf"
		echo "# Tivo S3 Emulator" >> $MOUNTPOINT/etc/tclient.conf
		echo "140::hd.oztivo.net:8000::::" >> $MOUNTPOINT/etc/tclient.conf
	else
		info "DailConfig 140 already exists"
	fi
	grep "^150::" $MOUNTPOINT/etc/tclient.conf >> ${LOGFILE}
	if [ $? -ne 0 ]; then
		info "Adding DailConfig 150 to tclient.conf"
		echo "# Tivo S3 development Emulator" >> $MOUNTPOINT/etc/tclient.conf
		echo "150::devhd.oztivo.net:8000::::" >> $MOUNTPOINT/etc/tclient.conf
	else
		info "DailConfig 150 already exists"
	fi
	info "Setting up bootpage Parameters=$params"
	root=$($BOOTPAGE -p $TIVO_DRIVE | grep -o "root=/dev/hda.")
	params="$root dsscon=true console=1,115200 hacks=telnet,ssh,$tmenable,serial upgradesoftware=false"
	$BOOTPAGE -P "$params" $TIVO_DRIVE || fail "Could not update boot page"
	$BOOTPAGE -p $TIVO_DRIVE >> $LOGFILE
	m=$(cat <<-EOF
	The following hacks have been installed:  ssh, telnet, BASH on serial, $tmenable 
	 \ndial code 140 for emulator added to tclient.conf
	 \nFiles for mfsftp, tivoweb and bufferhack copied to /hack .These will need additional configuration to use
	${PATCH_MSG}

	You can now replace the disk in your TiVo
	EOF
	)
	msg "$m"
}

log "START: $VERSION"
check_prereq
check_root || do_exit "Need root"
welcome || do_exit "User cancelled" 
while [ x${TIVO_DRIVE} = "x" ]; do
	select_drive || do_exit "User cancelled"
	check_drive ${DISK} 
done
confirm_drive || do_exit "User cancelled" 
patch_kernel
install_hacks
cleanup

# End of script
# vi: nobackup
